#define INITGUID
#include <windows.h>
#include <psapi.h>
#include <ddraw.h>
#include <stdio.h>
#include "ati3dcif.h"
#include "ctx.h"

HINSTANCE hInstance;

//FILE* flog = 0;

extern __declspec(dllimport) Ctx ctx;

C3D_EPRIM ptype = C3D_EPRIM_TRI;
C3D_EVERTEX vtype = C3D_EV_VTCF;
PDWORD palette = 0;


extern "C" BOOL WINAPI DllMain(HINSTANCE hInst, DWORD dwReason, LPVOID lpReserved)
{
	switch (dwReason)
	{
	case DLL_PROCESS_ATTACH:
		//flog = fopen("cif.log", "wt");
		hInstance = hInst;
		break;
	case DLL_PROCESS_DETACH:
		//fclose(flog);
		break;
	}
	return TRUE;
}


#define EXPORT(function_name, return_type, parameters) \
extern "C" __declspec(dllexport) return_type WINAPI function_name parameters; \
extern "C" __declspec(dllexport) void* function_name##_lib = function_name; \
extern "C" __declspec(dllexport) return_type WINAPI function_name parameters


EXPORT (ATI3DCIF_Init, C3D_EC, (void))
{
	return C3D_EC_OK;
}


EXPORT (ATI3DCIF_Term, C3D_EC, (void))
{
	//fprintf(flog, "term\n");
	return C3D_EC_OK;
}


EXPORT (ATI3DCIF_GetInfo, C3D_EC, (PC3D_3DCIFINFO p3DCIFInfo))
{
	//fprintf(flog, "GetInfo\n");
	if (p3DCIFInfo->u32Size < sizeof(C3D_3DCIFINFO)) return C3D_EC_BADPARAM;
    p3DCIFInfo->u32FrameBuffBase = 0;    // Host pointer to frame buffer base
    p3DCIFInfo->u32OffScreenHeap = 0;    // Host pointer to offscreen heap
    p3DCIFInfo->u32OffScreenSize = 4<<20;    // Size of offscreen heap
    p3DCIFInfo->u32TotalRAM = 16<<20;         // Total amount of RAM on the card
    p3DCIFInfo->u32ASICID = 1;           // ASIC Id. code
    p3DCIFInfo->u32ASICRevision = 1;     // ASIC revision
    p3DCIFInfo->u32CIFCaps1 = C3D_CAPS1_Z_BUFFER | C3D_CAPS1_CI8_TMAP;
    p3DCIFInfo->u32CIFCaps2 = C3D_CAPS2_TEXTURE_CLAMP | C3D_CAPS2_DESTINATION_ALPHA_BLEND;
    p3DCIFInfo->u32CIFCaps3 = 0;          // ATI3DCIF module capabilities (see flags)
    p3DCIFInfo->u32CIFCaps4 = 0;          // ATI3DCIF module capabilities (see flags)
    p3DCIFInfo->u32CIFCaps5 = 0;          // ATI3DCIF module capabilities (see flags)
	return C3D_EC_OK;
}


PDWORD buff;

EXPORT (ATI3DCIF_TextureReg, C3D_EC, (C3D_PTMAP pTex, C3D_PHTX pHandle))
{
	//fprintf(flog, "TextureReg fmt=%d, mip=%d, map0=%x, map1=%x\n", pTex->eTexFormat, pTex->bMipMap, pTex->apvLevels[0], pTex->apvLevels[1]);
	if (pTex->u32Size + 4 < sizeof(*pTex)) return C3D_EC_BADPARAM;

	// pro mipmapy je nutna konverze na RGB pTex->apvLevels[0]

	if (!buff) buff = new DWORD[0x10000];

	int size = 1l << (pTex->u32MaxMapXSizeLg2 + pTex->u32MaxMapYSizeLg2);
	PBYTE src = (PBYTE) pTex->apvLevels[0];

	switch (pTex->eTexFormat)
	{
	case C3D_ETF_CI8:
		for (int i=0; i<size; i++)
		{
			DWORD c = palette[src[i]];
			buff[i] = (c ^ pTex->clrTexChromaKey.u32All) & 0xffffff ? c | 0xff000000 : 0;
		}
		break;
	}

	GLuint id;
	glGenTextures(1, &id);
	glBindTexture(GL_TEXTURE_2D, id);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 4,
		1 << pTex->u32MaxMapXSizeLg2, 1 << pTex->u32MaxMapYSizeLg2,
		GL_RGBA, GL_UNSIGNED_BYTE, buff);

	*pHandle = (C3D_HTX) id;
	return C3D_EC_OK;
}


EXPORT (ATI3DCIF_TextureUnreg, C3D_EC, (C3D_HTX htxToUnreg))
{
	//fprintf(flog, "TextureUnreg\n");
	glDeleteTextures(1, (GLuint*) &htxToUnreg);
	return C3D_EC_OK;
}


EXPORT (ATI3DCIF_TexturePaletteCreate, C3D_EC, (C3D_ECI_TMAP_TYPE epalette, void* pPalette, C3D_PHTXPAL phtpalCreated))
{
	palette = (PDWORD) pPalette;
	//fprintf(flog, "TexturePaletteCreate\n");
	return C3D_EC_OK;
}


EXPORT (ATI3DCIF_TexturePaletteDestroy, C3D_EC, (C3D_HTXPAL htxpalToDestroy))
{
	//fprintf(flog, "TexturePaletteDestroy\n");
	return C3D_EC_OK;
}


EXPORT (ATI3DCIF_TexturePaletteAnimate, C3D_EC, (C3D_HTXPAL htxpalToAnimate,DWORD u32StartIndex,DWORD u32NumEntries,C3D_PPALETTENTRY pclrPalette))
{
	//fprintf(flog, "TexturePaletteAnimate\n");
	return C3D_EC_OK;
}


EXPORT (ATI3DCIF_ContextCreate, C3D_HRC, (void))
{
	//fprintf(flog, "ContextCreate\n");
	return "OK";
}


EXPORT (ATI3DCIF_ContextDestroy, C3D_EC, (C3D_HRC ctx))
{
	//fprintf(flog, "ContextDestroy\n");
	return C3D_EC_OK;
}


GLenum depthTab[] = {
	GL_NEVER,
    GL_NEVER,
    GL_LESS,
    GL_LEQUAL,
    GL_EQUAL,
    GL_GEQUAL,
    GL_GREATER,
    GL_NOTEQUAL,
    GL_ALWAYS };


EXPORT (ATI3DCIF_ContextSetState, C3D_EC, (C3D_HRC hrc, C3D_ERSID eRStateID, C3D_PRSDATA pRStateData))
{
	//fprintf(flog, "ContextSetState\n");
	switch (eRStateID)
	{
	case C3D_ERS_SOLID_CLR:{
		C3D_PCOLOR c = (C3D_PCOLOR) pRStateData;
		glColor3ub(c->r, c->g, c->b);
		break;}
	case C3D_ERS_TMAP_EN:
		if (ctx.texon = *(PBOOL) pRStateData) glEnable(GL_TEXTURE_2D);
		else glDisable(GL_TEXTURE_2D);
		break;
	case C3D_ERS_TMAP_SELECT:
		glBindTexture(GL_TEXTURE_2D, *(GLuint*) pRStateData);
		break;
	case C3D_ERS_PRIM_TYPE:
		ptype = *(C3D_PEPRIM) pRStateData;
		break;
	case C3D_ERS_VERTEX_TYPE:
		vtype = *(C3D_PEVERTEX) pRStateData;
		break;
	case C3D_ERS_Z_MODE:{
		DWORD mode = *(PDWORD) pRStateData;
		glDepthMask(ctx.depthmask = mode >= C3D_EZMODE_TESTON_WRITEZ ? GL_TRUE : GL_FALSE);
		if (ctx.depthon = mode > C3D_EZMODE_TESTON) glEnable(GL_DEPTH_TEST);
		else glDisable(GL_DEPTH_TEST);
		break;}
	case C3D_ERS_Z_CMP_FNC:{
		DWORD fn = *(PDWORD) pRStateData;
		//if (fn < C3D_EZCMP_MAX) glDepthFunc(depthTab[fn]);
		break;}
	}
	return C3D_EC_OK;
}


EXPORT (ATI3DCIF_RenderBegin, C3D_EC, (C3D_HRC hrc))
{
	//fprintf(flog, "RenderBegin\n");
	return C3D_EC_OK;
}


EXPORT (ATI3DCIF_RenderEnd, C3D_EC, (void))
{
	//fprintf(flog, "RenderEnd\n");
	return C3D_EC_OK;
}


EXPORT (ATI3DCIF_RenderSwitch, C3D_EC, (C3D_HRC hrc))
{
	//fprintf(flog, "RenderSwitch\n");
	ATI3DCIF_RenderEnd();
	ATI3DCIF_RenderBegin(hrc);
	return C3D_EC_OK;
}


template<class T> void RenderPrimStrip(T* v, int count)
{
	for (; --count >= 0; ++v)
	{
		if (ctx.texon) glTexCoord4f(v->s, v->t, 0, v->w);
		glColor3f(v->r / 255.0, v->g / 255.0, v->b / 255.0);
		glVertex3f(v->x, v->y, v->z);
	}
}

template<class T> void RenderPrimList(T** pv, int count)
{
	for (; --count >= 0; ++pv)
	{
		T* v = *pv;
		if (ctx.texon) glTexCoord4f(v->s, v->t, 0, v->w);
		glColor4f(v->r / 255.0, v->g / 255.0, v->b / 255.0, 1);
		glVertex3f(v->x, v->y, v->z);
	}
}

EXPORT (ATI3DCIF_RenderPrimStrip, C3D_EC, (C3D_VSTRIP verts, int count))
{
	//fprintf(flog, "RenderPrimStrip\n");
	switch (ptype)
	{
	case C3D_EPRIM_TRI:  glBegin(GL_TRIANGLE_STRIP); break;
	case C3D_EPRIM_QUAD: glBegin(GL_QUAD_STRIP); break;
	case C3D_EPRIM_LINE: glBegin(GL_LINE_STRIP); break;
	default: return C3D_EC_OK;
	}
	switch (vtype)
	{
	case C3D_EV_VTCF:     RenderPrimStrip((C3D_VTCF    *) verts, count); break;
	case C3D_EV_TLVERTEX: RenderPrimStrip((C3D_TLVERTEX*) verts, count); break;
	}
	glEnd();
	return C3D_EC_OK;
}


EXPORT (ATI3DCIF_RenderPrimList, C3D_EC, (C3D_VLIST verts, int count))
{
	//fprintf(flog, "RenderPrimList %d\n", ptype);
	switch (ptype)
	{
	case C3D_EPRIM_TRI:  glBegin(GL_TRIANGLES); break;
	case C3D_EPRIM_QUAD: glBegin(GL_QUADS); break;
	case C3D_EPRIM_LINE: glBegin(GL_LINES); break;
	default: return C3D_EC_OK;
	}
	switch (vtype)
	{
	case C3D_EV_VTCF:     RenderPrimList((C3D_VTCF    **) verts, count); break;
	case C3D_EV_TLVERTEX: RenderPrimList((C3D_TLVERTEX**) verts, count); break;
	}
	glEnd();
	return C3D_EC_OK;
}
